"""Tests for houdini_logging_tools.adapters.loggeradapter module."""

# Standard Library
import logging

# Third Party
import pytest

# Houdini Logging Tools
import houdini_logging_tools.adapters.loggeradapter

# Houdini
import hou

# Fixtures


@pytest.fixture
def test_adapter():
    """Fixture to provide a minimally set up HoudiniLoggerAdapter."""
    logger = logging.getLogger("test_logger")

    adapter = houdini_logging_tools.adapters.loggeradapter.HoudiniLoggerAdapter(logger)

    yield adapter


# Tests


class Test_HoudiniLoggerAdapter:
    """Test houdini_logging_tools.adapters.loggeradapter.HoudiniLoggerAdapter."""

    def test___init__(self, mocker):
        """Test object initialization."""
        test_logger = logging.getLogger("test_logger")
        extra = {"extra": "value"}

        mock_dialog = mocker.MagicMock(spec=bool)
        mock_node = mocker.MagicMock(spec=hou.Node)
        mock_status_bar = mocker.MagicMock(spec=bool)

        log = houdini_logging_tools.adapters.loggeradapter.HoudiniLoggerAdapter(
            test_logger, mock_node, mock_dialog, mock_status_bar, extra=extra
        )

        assert log.logger == test_logger
        assert log.extra == extra

        assert log._dialog == mock_dialog
        assert log._node == mock_node
        assert log._status_bar == mock_status_bar

    # Class Methods

    def test_from_name(self):
        """Test HoudiniLoggerAdapter.from_name()."""
        result = (
            houdini_logging_tools.adapters.loggeradapter.HoudiniLoggerAdapter.from_name(
                "test_name"
            )
        )

        assert result.logger.name == "test_name"

        result = (
            houdini_logging_tools.adapters.loggeradapter.HoudiniLoggerAdapter.from_name(
                "test_name",
                node=hou.node("/obj"),
                dialog=True,
                status_bar=True,
                extra={"foo": "bar"},
            )
        )

        assert result.logger.name == "test_name"
        assert result.node == hou.node("/obj")
        assert result.dialog
        assert result.status_bar
        assert result.extra == {"foo": "bar"}

    # Properties

    def test_dialog(self, test_adapter):
        """Test HoudiniLoggerAdapter.dialog."""
        assert not test_adapter.dialog

        test_adapter.dialog = True
        assert test_adapter._dialog

    def test_node(self, test_adapter):
        """Test HoudiniLoggerAdapter.node."""
        assert test_adapter.node is None

        test_node = hou.node("/obj")
        test_adapter.node = test_node
        assert test_adapter._node == test_node

    def test_status_bar(self, test_adapter):
        """Test HoudiniLoggerAdapter.status_bar."""
        assert not test_adapter.status_bar

        test_adapter.status_bar = True
        assert test_adapter._status_bar

    # Methods

    def test_process__node_arg(self, test_adapter):
        """Test HoudiniLoggerAdapter.process() when passing a node."""
        kwargs = {"extra": {"node": hou.node("/out")}}

        test_adapter.node = hou.node("/obj")

        result = test_adapter.process("test logger message", kwargs)

        assert result == ("/out - test logger message", kwargs)

    def test_process__node_property(self, test_adapter):
        """Test HoudiniLoggerAdapter.process() when using the node property."""
        kwargs = {"extra": {}}

        test_adapter.node = hou.node("/obj")

        result = test_adapter.process("test logger message", kwargs)

        assert result == ("/obj - test logger message", kwargs)

    def test_process__ui_passed_no_severity_no_title(
        self,
        mocker,
        test_adapter,
        mock_hou_ui,
        set_ui_available,
    ):
        """Test HoudiniLoggerAdapter.process() when passing 'dialog' and 'status_bar' via the extra dict."""
        mock_message = mocker.MagicMock(spec=str)

        kwargs = {"extra": {"dialog": True, "status_bar": True}}

        result = test_adapter.process(mock_message, kwargs)

        assert result == (mock_message, kwargs)

        mock_hou_ui.displayMessage.assert_called_with(
            mock_message, severity=hou.severityType.Message, title=None
        )
        mock_hou_ui.setStatusMessage.assert_called_with(
            mock_message, severity=hou.severityType.Message
        )

    def test_process__ui_properties_with_severity_and_title(
        self, mocker, test_adapter, mock_hou_ui, set_ui_available
    ):
        """Test HoudiniLoggerAdapter.process() passing 'dialog' and 'status_bar' via properties with a severity
        and title."""
        test_adapter.dialog = True
        test_adapter.status_bar = True

        mock_message = mocker.MagicMock(spec=str)
        mock_title = mocker.MagicMock(spec=str)

        kwargs = {"extra": {"severity": hou.severityType.Error, "title": mock_title}}

        result = test_adapter.process(mock_message, kwargs)

        assert result == (mock_message, kwargs)

        mock_hou_ui.displayMessage.assert_called_with(
            mock_message, severity=hou.severityType.Error, title=mock_title
        )
        mock_hou_ui.setStatusMessage.assert_called_with(
            mock_message, severity=hou.severityType.Error
        )

    @pytest.mark.parametrize("pass_dialog", (False, True))
    def test_process__message_args(
        self, test_adapter, mock_hou_ui, set_ui_available, pass_dialog
    ):
        """Test HoudiniLoggerAdapter.process() passing 'message_args'."""
        test_adapter.dialog = pass_dialog

        kwargs = {"extra": {"message_args": ("extra", 3)}}

        result = test_adapter.process("test logger message %s %d", kwargs)

        assert result == ("test logger message %s %d", kwargs)

        if pass_dialog:
            mock_hou_ui.displayMessage.assert_called_with(
                "test logger message extra 3",
                severity=hou.severityType.Message,
                title=None,
            )

    def test_process__message_args_no_display(self, test_adapter, mock_hou_ui):
        """Test HoudiniLoggerAdapter.process() passing 'message_args' but not displaying them."""
        kwargs = {"extra": {"message_args": ("extra", 3)}}

        result = test_adapter.process("test logger message %s %d", kwargs)

        assert result == ("test logger message %s %d", kwargs)

        mock_hou_ui.displayMessage.assert_not_called()

    def test_process__no_extra(self, mocker, test_adapter):
        """Test HoudiniLoggerAdapter.process() when passing an empty kwargs dict."""
        mock_message = mocker.MagicMock(spec=str)

        kwargs = {}

        result = test_adapter.process(mock_message, kwargs)

        assert result == (mock_message, kwargs)

    @pytest.mark.parametrize(
        "level, severity, num_message_args, passed_kwargs",
        [
            ("info", hou.severityType.ImportantMessage, 0, {"status_bar": True}),
            ("warning", hou.severityType.Warning, 1, {"node": hou.node("/obj")}),
            ("error", hou.severityType.Error, 2, {"dialog": True}),
            ("critical", hou.severityType.Error, 1, {"title": "A Title"}),
            ("debug", hou.severityType.Message, 1, {"stacklevel": 3}),
            ("exception", hou.severityType.Error, 1, {}),
        ],
    )
    def test_calls(
        self, mocker, test_adapter, level, severity, num_message_args, passed_kwargs
    ):
        """Test the various log calls.

        This helps to test the _wrap_logger functionality and that the wrapping occurred as expected.
        """
        mock_logger = mocker.MagicMock(spec=logging.Logger)
        mocker.patch.object(test_adapter, "logger", mock_logger)

        mock_msg = mocker.MagicMock(spec=str)

        message_args = tuple(
            [mocker.MagicMock(spec=str) for i in range(num_message_args)]
        )

        kwargs = {
            "foo": 3,  # A dummy extra kwarg our calling code does not care about
        }
        kwargs.update(passed_kwargs)

        # We're going to mock process() call to determine whether all our wrapper logic
        # runs and passes the expected data for the actual log call.
        mock_process = mocker.patch(
            "houdini_logging_tools.adapters.loggeradapter.HoudiniLoggerAdapter.process",
            return_value=(mocker.MagicMock(spec=str), kwargs),
        )

        expected_kwargs = {
            "foo": 3,
            "extra": {
                "severity": severity,
            },
            "stacklevel": passed_kwargs.get("stacklevel", 4),
        }

        # If there were any extra message args we expect them to have been added
        # to the extra dict.
        if num_message_args:
            expected_kwargs["extra"]["message_args"] = message_args

        for arg, value in passed_kwargs.items():
            if arg in ("node", "dialog", "status_bar", "title"):
                expected_kwargs["extra"][arg] = value

        # If logging an exception, ensure that exc_info=True is passed.
        if level == "exception":
            expected_kwargs["exc_info"] = True

        getattr(test_adapter, level)(mock_msg, *message_args, **kwargs)

        mock_process.assert_called_with(mock_msg, expected_kwargs)



"""Tests for houdini_logging_tools.handlers.shellio module."""

# Standard Library
import logging

# Third Party
import pytest

# Houdini Logging Tools
import houdini_logging_tools.handlers.shellio

# Houdini
import hou

# Fixtures


@pytest.fixture
def init_handler(mocker):
    """Fixture to initialize a handler."""
    mocker.patch.object(
        houdini_logging_tools.handlers.shellio.PythonShellHandler,
        "__init__",
        lambda x, y: None,
    )

    def _create():
        return houdini_logging_tools.handlers.shellio.PythonShellHandler(None)

    return _create


# Tests


class Test_PythonShellHandler:
    """Test houdini_logging_tools.handlers.shellio.PythonShellHandler object."""

    def test_emit__shellio(self, init_handler, mocker):
        """Test when sys.stdout is an instance of hou.ShellIO."""
        mock_format = mocker.patch.object(
            houdini_logging_tools.handlers.shellio.PythonShellHandler, "format"
        )

        mock_message = mocker.MagicMock(spec=str)

        mock_format.return_value = mock_message

        mock_record = mocker.MagicMock(spec=logging.LogRecord)
        inst = init_handler()

        mock_stream = mocker.patch("sys.stdout", spec=hou.ShellIO)
        inst.emit(mock_record)

        mock_format.assert_called_with(mock_record)

        calls = [mocker.call(mock_format.return_value), mocker.call("\n")]

        mock_stream.write.assert_has_calls(calls)
        mock_stream.flush.assert_called()

    def test_emit__not_shellio(self, init_handler, mocker):
        """Test when sys.stdout is not an instance of hou.ShellIO."""
        mock_format = mocker.patch.object(
            houdini_logging_tools.handlers.shellio.PythonShellHandler, "format"
        )

        mock_message = mocker.MagicMock(spec=str)

        mock_format.return_value = mock_message

        mock_record = mocker.MagicMock(spec=logging.LogRecord)
        inst = init_handler()

        inst.emit(mock_record)

        mock_format.assert_not_called()

    def test_emit__keyboardinterrupt(self, init_handler, mocker):
        """Test when KeyboardInterrupt is raised."""
        mock_format = mocker.patch.object(
            houdini_logging_tools.handlers.shellio.PythonShellHandler, "format"
        )
        mock_format.side_effect = KeyboardInterrupt

        mock_record = mocker.MagicMock(spec=logging.LogRecord)
        inst = init_handler()

        mocker.patch("sys.stdout", spec=hou.ShellIO)

        with pytest.raises(KeyboardInterrupt):
            inst.emit(mock_record)

        mock_format.assert_called_with(mock_record)

    def test_emit__systemexit(self, init_handler, mocker):
        """Test when SystemExit is raised."""
        mock_format = mocker.patch.object(
            houdini_logging_tools.handlers.shellio.PythonShellHandler, "format"
        )
        mock_format.side_effect = SystemExit

        mock_record = mocker.MagicMock(spec=logging.LogRecord)
        inst = init_handler()

        mocker.patch("sys.stdout", spec=hou.ShellIO)

        with pytest.raises(SystemExit):
            inst.emit(mock_record)

        mock_format.assert_called_with(mock_record)

    def test_emit__generic_exception(self, init_handler, mocker):
        """Test when a generic exception is raised."""
        mock_format = mocker.patch.object(
            houdini_logging_tools.handlers.shellio.PythonShellHandler, "format"
        )
        mock_format.side_effect = Exception

        mock_handle = mocker.patch.object(
            houdini_logging_tools.handlers.shellio.PythonShellHandler, "handleError"
        )

        mock_record = mocker.MagicMock(spec=logging.LogRecord)
        inst = init_handler()

        mocker.patch("sys.stdout", spec=hou.ShellIO)

        inst.emit(mock_record)

        mock_format.assert_called_with(mock_record)
        mock_handle.assert_called_with(mock_record)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
sys.path.insert(0, os.path.abspath("../src"))

project = 'houdini-logging-tools'
copyright = '2023, Graham Thompson'
author = 'Graham Thompson'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_rtd_theme",
    "sphinx_copybutton",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    "hou": ("https://www.sidefx.com/docs/houdini/hom/hou", "objects_hou.inv"),
}

autodoc_mock_imports = ["hou"]
autodoc_member_order = "bysource"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



"""Custom logging stream handler which writes to Houdini Python Shell panels."""

# Standard Library
import logging
import sys

# Houdini
import hou

# Classes


class PythonShellHandler(logging.StreamHandler):
    """Custom stream handler which outputs to the interactive Python shell
    when it is open.

    Houdini will redirect sys.stdout to be an instance of hou.ShellIO when there
    is a Python Shell panel active and displayed.  This handler works by checking
    that sys.stdout is a hou.ShellIO and writes output to it accordingly.  If
    it is not, no output will be written.

    """

    # Methods

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log message.

        Args:
            record:
                The log record to emit.

        """
        try:
            # Get the current stdout stream. Houdini will muck around with
            # this depending on whether a PythonShell panel is open.
            stream = sys.stdout

            # If the stream is really an output to a Python Shell then we know
            # that we want to write the message to it. Otherwise, a panel isn't
            # open, so we don't have anything to write to.
            if isinstance(stream, hou.ShellIO):
                # Format the message
                msg = self.format(record)

                stream.write(msg)
                stream.write("\n")
                stream.flush()

        # Re-raise these as we don't want to actually handle them.
        except (KeyboardInterrupt, SystemExit) as inst:
            raise inst

        # Otherwise handle the error.
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)



"""Custom logging adapter for Houdini."""

# Future
from __future__ import annotations

# Standard Library
import logging
from functools import wraps
from typing import Any, Callable, Optional, Tuple, Type

# Houdini
import hou

# Globals

# Call kwargs that should be moved into the extra data passed to process().
_KWARGS_TO_EXTRA_KEYS = (
    "node",
    "dialog",
    "status_bar",
    "title",
)

# Logging functions to wrap, and a corresponding severity for popup
# messages.
_TO_WRAP = {
    "critical": hou.severityType.Error,
    "debug": hou.severityType.Message,
    "error": hou.severityType.Error,
    "exception": hou.severityType.Error,
    "info": hou.severityType.ImportantMessage,
    "warning": hou.severityType.Warning,
}


# Classes


class HoudiniLoggerAdapter(logging.LoggerAdapter):
    """Custom LoggerAdapter for Houdini that allows automated addition of node
    paths and log display in dialogs, status bar, etc.  Also allows for
    automated notification.

    Args:
        base_logger:
            The base package logger.
        node:
            Optional node for prefixing messages with the path.
        dialog:
            Whether to always utilize the dialog option.
        status_bar:
            Whether to always utilize the dialog option.
        extra:
            Extra args to use to generate log messages.
    """

    def __init__(
        self,
        base_logger: logging.Logger,
        node: hou.Node = None,
        dialog: bool = False,
        status_bar: bool = False,
        extra: Optional[dict] = None,
    ) -> None:
        extra = extra if extra else {}

        super().__init__(base_logger, extra)

        self._dialog = dialog
        self._node = node
        self._status_bar = status_bar

    def __new__(
        cls: Type[HoudiniLoggerAdapter],  # pylint: disable=unused-argument
        *args: Any,
        **kwargs: Any,
    ) -> HoudiniLoggerAdapter:  # pragma: no cover
        """Overridden __new__ that will wrap logging methods with custom function."""
        inst = super(HoudiniLoggerAdapter, cls).__new__(cls)

        # We want to wrap various log calls to process args and set severities.
        for key, severity in _TO_WRAP.items():
            if hasattr(inst, key):
                attr = getattr(inst, key)

                if callable(attr):
                    wrapped = _wrap_logger(attr, severity)

                    setattr(inst, key, wrapped)

        return inst

    # Class Methods

    @classmethod
    def from_name(
        cls,
        name: str,
        node: hou.Node = None,
        dialog: bool = False,
        status_bar: bool = False,
        extra: Optional[dict] = None,
    ) -> HoudiniLoggerAdapter:
        """Create a new HoudiniLoggerAdapter from a name.

        This is a convenience function around the following code:

        >>> base_log = logging.getLogger(name)
        >>> logger = HoudiniLoggerAdapter(base_log)

        Args:
            name:
                The name of the logger to use.
            node:
                Optional node for prefixing messages with the path.
            dialog:
                Whether to always utilize the dialog option.
            status_bar:
                Whether to always utilize the dialog option.
            extra:
                Extra args to use to generate log messages.

        Returns:
            An adapter wrapping a logger of the passed name.

        """
        # Create a base logger
        base_logger = logging.getLogger(name)

        return cls(
            base_logger, node=node, dialog=dialog, status_bar=status_bar, extra=extra
        )

    # Properties

    @property
    def dialog(self) -> bool:
        """Whether the dialog will be displayed."""
        return self._dialog

    @dialog.setter
    def dialog(self, dialog: bool) -> None:
        self._dialog = dialog

    # --------------------------------------------------------------------------

    @property
    def node(self) -> Optional[hou.Node]:
        """A node the logger is associated with."""
        return self._node

    @node.setter
    def node(self, node: Optional[hou.Node]) -> None:
        self._node = node

    # --------------------------------------------------------------------------

    @property
    def status_bar(self) -> bool:
        """Whether the message will be logged to the status bar."""
        return self._status_bar

    @status_bar.setter
    def status_bar(self, status_bar: bool) -> None:
        self._status_bar = status_bar

    # Methods

    def process(self, msg: str, kwargs: Any) -> Tuple[str, Any]:
        """Override process() function to possibly insert a node path or to
        display a dialog with the log message before being passed to regular
        logging output.

        Args:
            msg:
                The log message.
            kwargs:
                kwargs dict.

        Returns:
            The message and updated kwargs.

        """
        extra: dict = self.extra  # type: ignore

        if "extra" in kwargs:
            extra.update(kwargs["extra"])

            node = extra.pop("node", self.node)

            # Prepend the message with the node path.
            if node is not None:
                msg = f"{node.path()} - {msg}"

            dialog = extra.pop("dialog", self.dialog)
            status_bar = extra.pop("status_bar", self.status_bar)

            if hou.isUIAvailable():
                # Copy of the message for our display.
                houdini_message = msg

                # If we have message args we need to format the message with them.
                if "message_args" in extra:
                    houdini_message = houdini_message % extra["message_args"]

                severity = extra.pop("severity", hou.severityType.Message)

                # Display the message as a popup.
                if dialog:
                    title = extra.pop("title", None)

                    hou.ui.displayMessage(
                        houdini_message, severity=severity, title=title
                    )

                if status_bar:
                    hou.ui.setStatusMessage(houdini_message, severity=severity)

        kwargs["extra"] = extra

        return msg, kwargs


# Non-Public Functions


def _wrap_logger(func: Callable, severity: hou.severityType) -> Callable:
    """Function which wraps a logger method with custom code.

    Args:
        func:
            The callable to wrap.

        severity:
            The corresponding hou.severityType value.

    Returns:
        The wrapped function.

    """

    @wraps(func)
    def func_wrapper(*args: Any, **kwargs: Any) -> Any:
        # Get the extra dictionary, or an empty one if it doesn't exist.
        extra = kwargs.setdefault("extra", {})

        # Set the severity to our passed in value.
        extra["severity"] = severity

        for key in _KWARGS_TO_EXTRA_KEYS:
            if key in kwargs:
                extra[key] = kwargs.pop(key)

        # If there is more than one arg we want to pass them as extra data so that
        # we can use it to format the message for extra outputs.
        if len(args) > 1:
            extra["message_args"] = args[1:]

        if "stacklevel" not in kwargs:
            # Set stacklevel=4 so that the module/file/line reporting will represent
            # the calling point and not the function call inside the adapter.
            kwargs["stacklevel"] = 4

        return func(*args, **kwargs)

    return func_wrapper
